/*

Copyright (C) 2025 fijxu@nadeko.net

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, version 3.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <https://www.gnu.org/licenses/>.

*/

package main

import (
	"context"
	"encoding/json"
	"flag"
	"fmt"
	"log"
	"net"
	"os"
	"strings"
	"time"

	utils "git.nadeko.net/fijxu/simple-ddns-client/internal"
)

type Porkbun struct {
	ApiKey       string `json:"apiKey"`
	SecretApiKey string `json:"secretApiKey"`
	Domain       string `json:"domain"`
	Type         string `json:"type"`
	// rfc1035 defines: "TTL: positive values of a signed 32 bit number."
	Ttl  int32 `json:"ttl"`
	Data PorkbunData
}

type PorkbunData struct {
	SecretApiKey *string `json:"secretapikey"`
	ApiKey       *string `json:"apikey"`
	Ip           string  `json:"content"`
	Ttl          string  `json:"ttl"`
}

type Config struct {
	UpdateInterval int     `json:"updateInterval"`
	Provider       string  `json:"provider"`
	DnsServer      string  `json:"dnsServer"`
	PublicIpApi    string  `json:"publicIpApi"`
	Porkbun        Porkbun `json:"porkbun"`
}

var config *Config

// https://stackoverflow.com/questions/59889882/specifying-dns-server-for-lookup-in-go
// This will only work for looking up the IP address of the domain that we
// want to update. Other requests like HTTP ones will use the system DNS
// servers.
var resolver = &net.Resolver{
	PreferGo: true,
	Dial: func(ctx context.Context, network, address string) (net.Conn, error) {
		d := net.Dialer{
			Timeout: time.Millisecond * time.Duration(5000),
		}
		if config.DnsServer == "" {
			// TODO: This is called two times. idk why
			log.Print("\"dnsServer\" key is empty. Using default DNS server")
			return d.DialContext(ctx, network, address)
		}
		return d.DialContext(ctx, network, config.DnsServer)
	},
}

// TODO: Use UPNP, NAT-PMP or with PCP
func getIpAddress() string {
	req, err := utils.DoGetRequest(config.PublicIpApi)
	if err != nil {
		log.Print("Failed to retrieve public ip address from " + config.PublicIpApi + ": " + err.Error())
		return ""
	}
	return string(req)
}

func (p *Porkbun) updateIp() {
	dnsIp, err := resolver.LookupHost(context.Background(), p.Domain)
	p.Data.Ip = getIpAddress()
	if p.Data.Ip == "" {
		log.Print("Failed to retrieve current public IP address")
		return
	}

	if err != nil {
		log.Print("Failed to retrieve IP address for domain " + p.Domain + ": " + err.Error())
		return
	}
	if len(dnsIp) > 1 {
		log.Fatal("The record '" + p.Domain + "' can't contain more than 1 IP address!")
	}
	if string(dnsIp[0]) == p.Data.Ip {
		log.Print("No need to update the IP address of domain '" + p.Domain + "'. IP is already " + p.Data.Ip)
		return
	}
	log.Print("Current IP of the record '" + p.Domain + "' is " + dnsIp[0])

	// Porkbun uses a string for the `"ttl"` key.
	p.Data.Ttl = string(p.Ttl)
	p.Data.ApiKey = &p.ApiKey
	p.Data.SecretApiKey = &p.SecretApiKey
	domainParts := strings.Split(p.Domain, ".")
	// TODO: Support 4th level FQDNs. Ex: test.us.nadeko.net
	domain := domainParts[1] + "." + domainParts[2]
	subDomain := domainParts[0]

	data, _ := json.Marshal(&p.Data)
	queryUrl := domain + "/" + p.Type + "/" + subDomain

	res, err := utils.DoPostRequest("https://api.porkbun.com/api/json/v3/dns/editByNameType/"+queryUrl, data)
	if err != nil {
		log.Print("Failed to update public IP address: " + err.Error())
		return
	}

	// Porkbun returns `{"status":"SUCCESS"}` if the IP address has been
	// updated successfully
	if strings.Contains(string(res), "SUCCESS") {
		log.Print("IP Address updated to " + p.Data.Ip + " for domain " + p.Domain)
	} else {
		log.Print("ERROR: " + string(res))
	}
}

func (p *Config) updateIp() {
	switch p.Provider {
	case "porkbun":
		p.Porkbun.updateIp()
	}
}

func loadConfig(configPath string) *Config {
	var config = &Config{}

	file, err := os.ReadFile(configPath)
	if err != nil {
		log.Fatal("Failed to parse config file: " + err.Error() + ". Exiting")
	}

	err = json.Unmarshal(file, config)
	if err != nil {
		log.Fatal("Failed to parse config file: " + err.Error() + ". Exiting")
	}

	return config
}

func init() {
	var configPath string

	flag.StringVar(&configPath, "c", "/etc/simple-ddns-client/config.json", "config.json path")
	flag.Parse()

	config = loadConfig(configPath)

	if config.UpdateInterval <= 0 {
		log.Fatal("\"updateInterval\" has to be greater than 0")
	}
}

func main() {
	for {
		config.updateIp()
		log.Print("Sleeping for " + fmt.Sprint(config.UpdateInterval) + " seconds")
		time.Sleep(time.Duration(config.UpdateInterval) * time.Second)
	}
}
